@@ -188,9 +188,11 @@ enum network_transport {
     udp_transport
 };
 
-enum item_lock_types {
-    ITEM_LOCK_GRANULAR = 0,
-    ITEM_LOCK_GLOBAL
+enum pause_thread_types {
+    PAUSE_WORKER_THREADS = 0,
+    PAUSE_ALL_THREADS,
+    RESUME_ALL_THREADS,
+    RESUME_WORKER_THREADS
 };
 
 #define IS_UDP(x) (x == udp_transport)
@@ -389,7 +391,6 @@ typedef struct {
     struct thread_stats stats;  /* Stats generated by this thread */
     struct conn_queue *new_conn_queue; /* queue of new connections to handle */
     cache_t *suffix_cache;      /* suffix cache */
-    uint8_t item_lock_type;     /* use fine-grained or global item lock */
 } LIBEVENT_THREAD;
 
 typedef struct {
@@ -547,7 +548,7 @@ static inline int mutex_lock(pthread_mutex_t *mutex)
  * also #define-d to directly call the underlying code in singlethreaded mode.
  */
 
-void thread_init(int nthreads, struct event_base *main_base);
+void memcached_thread_init(int nthreads, struct event_base *main_base);
 int  dispatch_event_add(int thread, conn *c);
 void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags, int read_buffer_size, enum network_transport transport);
 
@@ -574,13 +575,11 @@ void  item_stats_sizes(ADD_STAT add_stats, void *c);
 void  item_unlink(item *it);
 void  item_update(item *it);
 
-void item_lock_global(void);
-void item_unlock_global(void);
 void item_lock(uint32_t hv);
 void *item_trylock(uint32_t hv);
 void item_trylock_unlock(void *arg);
 void item_unlock(uint32_t hv);
-void switch_item_lock_type(enum item_lock_types type);
+void pause_threads(enum pause_thread_types type);
 unsigned short refcount_incr(unsigned short *refcount);
 unsigned short refcount_decr(unsigned short *refcount);
 void STATS_LOCK(void);
